/*BEGIN_LEGAL 
Intel Open Source License 

Copyright (c) 2002-2011 Intel Corporation. All rights reserved.
 
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.  Redistributions
in binary form must reproduce the above copyright notice, this list of
conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.  Neither the name of
the Intel Corporation nor the names of its contributors may be used to
endorse or promote products derived from this software without
specific prior written permission.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE INTEL OR
ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
END_LEGAL */

/*! @file
 *  A simple exploit generator
 *  This program identifies overwrites of function return address
 *  in the program main module and returns the source of the taint
 *  A supporting python script of same name exists to execute this 
 *  tool and genereate a simple exploit that jumps to the longest 
 *  consecutive tainted location in memory and puts the target 
 *  buffer there.
 *  note that the heap / stack address delta may vary on different
 *  systems and you should change the script constant or use attach
 *  to work around the address delta problem.
 *  
 *  Limitation: taint propogated through MOV* instructions only
 */

#include "pin.H"
#include <iostream>
#include <fstream>
#include <map>
#include <stack>
#include <string.h>
#include "xed-iclass-enum.h"

/* ================================================================== */
// Global variables 
/* ================================================================== */

map<ADDRINT, ADDRINT> TaintedAddrs;		// tainted memory addresses
ADDRINT TaintedRegs[REG_LAST] = {0};	// tainted registers
std::ofstream out;						// output file

typedef struct
{
    ADDRINT address;
    ADDRINT value;
} pAddr;

stack<pAddr> protect;					//addresses to protect

/* ===================================================================== */
// Command line switches
/* ===================================================================== */
KNOB<string> KnobOutputFile(KNOB_MODE_WRITEONCE,  "pintool",
    "o", "exploit.out", "specify file name for the output file");


/* ===================================================================== */
// Utilities
/* ===================================================================== */

/*!
 *  Print out help message.
 */
INT32 Usage()
{
    cerr << "This tool follows the taint defined by the first argument to " << endl <<
            "the instumented program command line and outputs details to " << endl <<
            "a file when it identifies an overwrite of a return address " << endl << endl;

    cerr << KNOB_BASE::StringKnobSummary() << endl;

    return -1;
}

//Dump taint information to our log file
VOID DumpTaint()
{
    out << "## Begin Dump" << endl;
    map<ADDRINT, ADDRINT>::iterator it;
    for ( it=TaintedAddrs.begin() ; it != TaintedAddrs.end(); it++ )
    {
        out << "T " << (*it).first << " source " << (*it).second << endl;
    }
    out << "## End Dump" << endl;
}

/* ===================================================================== */
// Analysis routines
/* ===================================================================== */


// Save address to protect on entry to function
VOID RtnEntry(ADDRINT esp, ADDRINT addr)
{
    //we're at RTN_Entry to we're just after the call and before the 
    //function prolog, therefore the return address is at TOS
    pAddr tmp;
    tmp.address = esp; 
    tmp.value = *((ADDRINT *)esp);
    protect.push(tmp);
}

// check if return address was overwritten
VOID RtnExit(ADDRINT esp, ADDRINT addr)
{
    if (protect.empty())
    {
        out << "WARNING! protection list empty" << endl;
        return;
    }

    pAddr orig = protect.top();
    
    if (esp != orig.address)
    {
        out << "ESP: " << esp << " orig: " << orig.address << endl;
    }
    
    ADDRINT cur_val = (*((ADDRINT *)orig.address));
    if (orig.value != cur_val)
    {
        out << "Overwrite at: " << orig.address << " old value: " << orig.value 
            << " new value: " << cur_val << endl;
        for (int i=0; i<4; i++)
        {
            out << "Source of taint at: " << (orig.address + i) << " is: " 
                << TaintedAddrs[orig.address+i] << endl;
        }	

        out << "Dumping taint" << endl;
        DumpTaint();
    }

    protect.pop();
}


// This functions marks the contents of argv[1] as tainted
VOID MainAddTaint(unsigned int argc, char *argv[])
{
    if (argc != 2)
    {
        return;
    }

    int n = strlen(argv[1]);
    ADDRINT taint = (ADDRINT)argv[1];
    for (int i = 0; i < n; i++)
    {
        TaintedAddrs[taint + i] = i+1;
    }
}

// This function represents the case of a register copied to memory
void RegTaintMem(ADDRINT reg_r, ADDRINT mem_w)
{
    if (TaintedRegs[reg_r]) 
    {
        TaintedAddrs[mem_w] = TaintedRegs[reg_r];
    }
    else //reg not tainted --> mem not tainted
    {
        if (TaintedAddrs.count(mem_w)) // if mem is already not tainted nothing to do
        {
            TaintedAddrs.erase(mem_w);
        }
    }
}

// this function represents the case of a memory copied to register
void MemTaintReg(ADDRINT mem_r, ADDRINT reg_w, ADDRINT inst_addr)
{
    if (TaintedAddrs.count(mem_r)) //count is either 0 or 1 for set
    {
        TaintedRegs[reg_w] = TaintedAddrs[mem_r];
    }
    else //mem is clean -> reg is cleaned
    {
        TaintedRegs[reg_w] = 0;
    }
}

// this function represents the case of a reg copied to another reg
void RegTaintReg(ADDRINT reg_r, ADDRINT reg_w)
{
    TaintedRegs[reg_w] = TaintedRegs[reg_r];
}

// this function represents the case of an immediate copied to a register
void ImmedCleanReg(ADDRINT reg_w)
{
    TaintedRegs[reg_w] = 0;
}

// this function represent the case of an immediate copied to memory
void ImmedCleanMem(ADDRINT mem_w)
{
    if (TaintedAddrs.count(mem_w)) // if mem is already not tainted nothing to do
    {
        TaintedAddrs.erase(mem_w);
    }
}


/* ===================================================================== */
// Instrumentation callbacks & Helpers
/* ===================================================================== */

// True if the instruction has an immediate operand
// meant to be called only from instrumentation routines
bool INS_has_immed(INS ins)
{
    for (unsigned int i = 0; i < INS_OperandCount(ins); i++)
    {
        if (INS_OperandIsImmediate(ins, i))
        {
            return true;
        }
    }
    return false;
}

// returns the full name of the first register operand written
REG INS_get_write_reg(INS ins)
{
    for (unsigned int i = 0; i < INS_OperandCount(ins); i++)
    {
        if (INS_OperandIsReg(ins, i) && INS_OperandWritten(ins, i))
        {
            return REG_FullRegName(INS_OperandReg(ins, i));
        }
    }
    
    return REG_INVALID();
}

// returns the full name of the first register operand read
REG INS_get_read_reg(INS ins)
{
    for (unsigned int i = 0; i < INS_OperandCount(ins); i++)
    {
        if (INS_OperandIsReg(ins, i) && INS_OperandRead(ins, i))
        {
            return REG_FullRegName(INS_OperandReg(ins, i));
        }
    }
    
    return REG_INVALID();
}

/*!
 * This function checks for each instruction if it does a mov that can potentially
 * transfer taint and if true adds the approriate analysis routine to check 
 * and propogate taint at run-time if needed
 * This function is called every time a new trace is encountered.
 */
VOID Trace(TRACE trace, VOID *v)
{
  for (BBL bbl = TRACE_BblHead(trace); BBL_Valid(bbl); bbl = BBL_Next(bbl))
  {
    for (INS ins = BBL_InsHead(bbl); INS_Valid(ins); ins = INS_Next(ins))
    {
        if ( (INS_Opcode(ins) >= XED_ICLASS_MOV) && (INS_Opcode(ins) <= XED_ICLASS_MOVZX) )
        {
            if (INS_has_immed(ins))
            {
                if (INS_IsMemoryWrite(ins)) //immed -> mem
                {
                    INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)ImmedCleanMem,
                                            IARG_MEMORYOP_EA, 0,
                                            IARG_END);
                }
                else						//immed -> reg
                {
                    REG insreg = INS_get_write_reg(ins);
                    INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)ImmedCleanReg,
                                            IARG_ADDRINT, (ADDRINT)insreg,
                                            IARG_END);
                }
            }
            else if (INS_IsMemoryRead(ins)) //mem -> reg 
            {
                //in this case we call MemTaintReg to copy the taint if relevant
                REG insreg = INS_get_write_reg(ins);
                INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)MemTaintReg,
                                        IARG_MEMORYOP_EA, 0,
                                        IARG_ADDRINT, (ADDRINT)insreg, IARG_INST_PTR,
                                        IARG_END);

            }
            else if (INS_IsMemoryWrite(ins)) //reg -> mem 
            {
                //in this case we call RegTaintMem to copy the taint if relevant
                REG insreg = INS_get_read_reg(ins); 
                INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)RegTaintMem,
                                        IARG_ADDRINT, (ADDRINT)insreg,
                                        IARG_MEMORYOP_EA, 0,
                                        IARG_END);
            }
            else if (INS_RegR(ins, 0) != REG_INVALID()) //reg -> reg
            {
                //in this case we call RegTaintReg
                REG Rreg = INS_get_read_reg(ins); 
                REG Wreg = INS_get_write_reg(ins); 
                INS_InsertCall(ins, IPOINT_BEFORE, (AFUNPTR)RegTaintReg,
                                        IARG_ADDRINT, (ADDRINT)Rreg,
                                        IARG_ADDRINT, (ADDRINT)Wreg,
                                        IARG_END);
            }
            else	//should never happen
            {
                out << "serious error?!\n" << endl;
            }
        } // IF opcode is a MOV
    }  // For INS
  }  // For BBL
} // VOID Trace


/*!
 * Routine instrumentaiton, called for every routine loaded
 * this function adds a call to MainAddTaint on the main function 
 */
VOID Routine(RTN rtn, VOID *v)
{
    RTN_Open(rtn);
    
    if (RTN_Name(rtn) == "main") //if this is the main function
    {
        RTN_InsertCall(rtn, IPOINT_BEFORE, (AFUNPTR)MainAddTaint,
                       IARG_FUNCARG_ENTRYPOINT_VALUE, 0,
                       IARG_FUNCARG_ENTRYPOINT_VALUE, 1,
                       IARG_END);
    }

    //find section and image
    SEC sec = RTN_Sec(rtn);
    IMG img = SEC_Img(sec);

    //if we're in the main executable insert retaddr overwrite checks
    if ( IMG_IsMainExecutable(img) && (SEC_Name(sec) == ".text") )
    {
        RTN_InsertCall(rtn, IPOINT_BEFORE,(AFUNPTR)RtnEntry, IARG_REG_VALUE, REG_ESP, IARG_INST_PTR, IARG_END);
        RTN_InsertCall(rtn, IPOINT_AFTER ,(AFUNPTR)RtnExit , IARG_REG_VALUE, REG_ESP, IARG_INST_PTR, IARG_END);
    }

    RTN_Close(rtn);	
}


/*!
 * Print out the taint analysis results.
 * This function is called when the application exits.
 */
VOID Fini(INT32 code, VOID *v)
{
    //DumpTaint(); //could be useful to turn on if you're debugging the tool
    out.close();
}



/*!
 * The main procedure of the tool.
 * This function is called when the application image is loaded but not yet started.
 * @param[in]   argc            total number of elements in the argv array
 * @param[in]   argv            array of command line arguments, 
 *                              including pin -t <toolname> -- ...
 */
int main(int argc, char *argv[])
{
    // Initialize PIN library. Print help message if -h(elp) is specified
    // in the command line or the command line is invalid 
    PIN_InitSymbols();

    if( PIN_Init(argc,argv) )
    {
        return Usage();
    }
    
    // Register function to be called to instrument traces
    TRACE_AddInstrumentFunction(Trace, 0);
    RTN_AddInstrumentFunction(Routine, 0);

    // Register function to be called when the application exits
    PIN_AddFiniFunction(Fini, 0);
    
    cerr <<  "===============================================" << endl;
    cerr <<  "This application is instrumented by MyPinTool" << endl;
    if (!KnobOutputFile.Value().empty()) 
    {
        cerr << "See file " << KnobOutputFile.Value() << " for analysis results" << endl;

        string fileName = KnobOutputFile.Value();
        out.open(fileName.c_str());
        out << hex;
    }
    cerr <<  "===============================================" << endl;

    // Start the program, never returns
    PIN_StartProgram();
    
    return 0;
}

/* ===================================================================== */
/* eof */
/* ===================================================================== */
