/*BEGIN_LEGAL 
Intel Open Source License 

Copyright (c) 2002-2011 Intel Corporation. All rights reserved.
 
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.  Redistributions
in binary form must reproduce the above copyright notice, this list of
conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.  Neither the name of
the Intel Corporation nor the names of its contributors may be used to
endorse or promote products derived from this software without
specific prior written permission.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE INTEL OR
ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
END_LEGAL */

/*
 *  This file contains a PIN tool for protecting function return addresses 
 *  from overwrites - applied only to functions in the main executable.
 */

#include <stdio.h>
#include "pin.H"
#include <stack>

typedef struct
{
    ADDRINT address;
    ADDRINT value;
} pAddr;

stack<pAddr> protect;	//addresses to protect

FILE * logfile;			//log file


// called at the end of the process execution
VOID Fini(INT32 code, VOID *v)
{
    fclose(logfile);
}

// Save address to protect on entry to function
VOID RtnEntry(ADDRINT esp, ADDRINT addr)
{
    pAddr tmp;
    tmp.address = esp;
    tmp.value = *((ADDRINT *)esp);
    protect.push(tmp);
}

// check if return address was overwritten
VOID RtnExit(ADDRINT esp, ADDRINT addr)
{
    if (protect.empty())
    {
        fprintf(logfile, "WARNING! protection list empty\n");
        return;
    }

    pAddr orig = protect.top();
    ADDRINT cur_val = (*((ADDRINT *)orig.address));
    if (orig.value != cur_val)
    {
        fprintf(logfile, "Overwrite at: %x old value: %x, new value: %x\n", 
                       orig.address, orig.value, cur_val );
        //we might want to dump the stack trace we have here for easier debugging
        //we can also restore the original value to prevent exploit
    }

    protect.pop();
}

//Called for every RTN, add calls to RtnEntry and RtnExit
VOID Routine(RTN rtn, VOID *v)
{
    RTN_Open(rtn);
    SEC sec = RTN_Sec(rtn);
    IMG img = SEC_Img(sec);

    if ( IMG_IsMainExecutable(img) && (SEC_Name(sec) == ".text") )
    {
        RTN_InsertCall(rtn, IPOINT_BEFORE,(AFUNPTR)RtnEntry, 
                        IARG_REG_VALUE, REG_ESP, 
                        IARG_INST_PTR, 
                        IARG_END);
        RTN_InsertCall(rtn, IPOINT_AFTER ,(AFUNPTR)RtnExit , 
                        IARG_REG_VALUE, REG_ESP, 
                        IARG_INST_PTR, 
                        IARG_END);
    }

    RTN_Close(rtn);
}


// Help message
INT32 Usage()
{
    PIN_ERROR( "This Pintool logs function return addresses in main module and reports modifications\n" 
              + KNOB_BASE::StringKnobSummary() + "\n");
    return -1;
}


// Tool main function - initialize and set instrumentation callbacks
int main(int argc, char *argv[])
{
    // initialize Pin + symbol processing
    PIN_InitSymbols();
    if (PIN_Init(argc, argv)) return Usage();

    // open logfile
    logfile = fopen("protection.out", "w");

    // set callbacks
    RTN_AddInstrumentFunction(Routine, 0);
    PIN_AddFiniFunction(Fini, 0);

    // Never returns
    PIN_StartProgram();
    
    return 0;
}
